@@ -194,7 +194,12 @@ int Detail(char *dev, int brief, int export, int test, char *homehost)
 				st->ss->export_detail_super(st);
 		} else {
 			struct map_ent *mp, *map = NULL;
+			char nbuf[64];
 			mp = map_by_devnum(&map, fd2devnum(fd));
+			if (mp) {
+				__fname_from_uuid(mp->uuid, 0, nbuf, ':');
+				printf("MD_UUID=%s\n", nbuf+5);
+			}
 			if (mp && mp->path &&
 			    strncmp(mp->path, "/dev/md/", 8) == 0)
 				printf("MD_DEVNAME=%s\n", mp->path+8);
@@ -207,11 +212,11 @@ int Detail(char *dev, int brief, int export, int test, char *homehost)
 		printf("ARRAY %s", dev);
 		if (brief > 1) {
 			if (array.raid_disks)
-				printf("level=%s num-devices=%d",
+				printf(" level=%s num-devices=%d",
 				       c?c:"-unknown-",
 				       array.raid_disks );
 			else
-				printf("level=container num-devices=%d",
+				printf(" level=container num-devices=%d",
 				       array.nr_disks);
 		}
 		if (container) {
@@ -320,6 +325,10 @@ int Detail(char *dev, int brief, int export, int test, char *homehost)
 			c = map_num(r5layout, array.layout);
 			printf("         Layout : %s\n", c?c:"-unknown-");
 		}
+		if (array.level == 6) {
+			c = map_num(r6layout, array.layout);
+			printf("         Layout : %s\n", c?c:"-unknown-");
+		}
 		if (array.level == 10) {
 			printf("         Layout :");
 			print_r10_layout(array.layout);
@@ -373,6 +382,11 @@ This is pretty boring
 					printf("     New Layout : %s\n",
 					       c?c:"-unknown-");
 				}
+				if (info.new_level == 6) {
+					char *c = map_num(r6layout, info.new_layout);
+					printf("     New Layout : %s\n",
+					       c?c:"-unknown-");
+				}
 				if (info.new_level == 10) {
 					printf("     New Layout : near=%d, %s=%d\n",
 					       info.new_layout&255,
@@ -536,6 +550,7 @@ This is pretty boring
 		    1, avail, avail_disks))
 		rv = 2;
 
+	free(disks);
 out:
 	close(fd);
 	return rv;
